# -*- coding: utf-8 -*-
from __future__ import annotations

import logging
import sys

import numpy as np
from astropy import units as u
from astropy.io import ascii
from astropy.io import fits
from astropy.units import Quantity
from scipy.interpolate import interp1d

from .configuration import get_config_file
from .filter import Filter

logger = logging.getLogger("milespy.magnitudes")

solar_ref_spec = get_config_file("sun_mod_001.fits")


class Magnitude(dict):
    def write(self, output=sys.stdout, format="basic", **kwargs):
        """
        Save the magnitude data in the requested format

        Any extra keyword parameters are passed to astropy.io.ascrii.write

        Parameters
        ----------
        output: str
            Output filename. Defaults to sys.stdout
        format : str
            Any of the available format specifier for astropy.io.ascii.write:
            https://docs.astropy.org/en/stable/api/astropy.io.ascii.write.html#astropy.io.ascii.write
        """
        from astropy.table import Table

        to_save = dict(self)
        for k in to_save:
            if np.isscalar(to_save[k]):
                to_save[k] = np.array([to_save[k]])

        tab = Table(data=to_save)
        ascii.write(tab, output, format=format, **kwargs)


def _load_zerofile(zeropoint):
    file = get_config_file("vega.sed")
    data = ascii.read(file, comment=r"\s*#")

    sed = {"wave": np.array(data["col1"]), "flux": np.array(data["col2"])}

    # If AB mags only need wavelength vector
    if zeropoint == "AB":
        sed["flux"] = 1.0 / np.power(sed["wave"], 2)
    elif zeropoint == "VEGA":
        # Normalizing the SED@ 5556.0\AA
        zp5556 = 3.44e-9  # erg cm^-2 s^-1 A^-1, Hayes 1985
        interp = interp1d(sed["wave"], sed["flux"])
        sed["flux"] *= zp5556 / interp(5556.0)

    return sed


# Pre-load the sed for the different zero points
zerosed = {"AB": _load_zerofile("AB"), "VEGA": _load_zerofile("VEGA")}


def compute_mags(
    wave: Quantity, flux: Quantity, filters: [Filter], zeropoint, sun=False
) -> Magnitude:
    """
    Compute the magnitudes given a spectra and a set of fluxes

    Core implementation of the magnitude computation, including the checks for
    the filter wavelength limits

    Parameters
    ----------
    wave : array
        Wavelength of the spectrum
    flux : array
        Spectrum
    filters : list[Filter]
        List of filters for which the magnitude is to be computed
    zeropoint : str
        Type of zero point. Either AB or VEGA
    sun : bool
        Flag to determine if the output is absolute magnitudes (True) or in
        flux (False) ??

    Returns
    -------
    Magnitude
    """
    # TODO: take into account the units
    flux = flux.to_value()
    wave = wave.to_value()
    # Defining some variables
    cvel = 2.99792458e18  # Speed of light in Angstron/sec
    dl = 1e-5  # 10 pc in Mpc, z=0; for absolute magnitudes
    if sun:
        cfact = -5.0 * np.log10(4.84e-6 / 10.0)  # for absolute magnitudes
    else:
        cfact = 5.0 * np.log10(1.7684e8 * dl)  # from lum[erg/s/A] to flux [erg/s/A/cm2]

    # Default is nan to mark an invalid range/filter
    outmags = Magnitude((f.name, np.full(flux.shape[:-1], np.nan)) for f in filters)

    # Computing the magnitude for each filter
    for filt in filters:
        # Finding the wavelength limits of the filters
        good = filt.wave > 0.0
        wlow = np.amin(filt.wave[good])
        whi = np.amax(filt.wave[good])

        # Selecting the relevant pixels in the input spectrum
        w = (wave >= wlow) & (wave <= whi)
        tmp_wave = wave[w]
        tmp_flux = flux[..., w]
        if (np.amin(wave) > wlow) or (np.amax(wave) < whi):
            logger.warning(
                f"Filter {filt.name} [{wlow},{whi}] is outside of"
                f"the spectral range [{np.amin(wave)}, {np.amax(wave)}]"
            )
            continue

        # Interpolate the filter response to data wavelength
        interp = interp1d(
            filt.wave[good],
            filt.trans[good],
        )
        response = interp(tmp_wave)

        # Calculating the magnitude in the desired system
        vega = np.interp(
            tmp_wave, zerosed[zeropoint]["wave"], zerosed[zeropoint]["flux"]
        )
        vega_f = np.trapz(vega * response, x=tmp_wave)

        f = np.trapz(tmp_flux * response, x=tmp_wave, axis=-1)
        mag = -2.5 * np.log10(f / vega_f)
        fmag = mag + cfact
        if zeropoint == "AB":
            fmag = fmag + 2.5 * np.log10(cvel) - 48.6  # oke & gunn 83

        outmags[filt.name] = fmag

    return outmags


def vacuum2air(wave_vac):
    """
    Converts wavelength from vacuum to air

    Parameters
    ----------
    array
        Wavelength in vacuum system

    Returns
    -------
    array
        Vector with wavelength in air system

    """

    wave_air = wave_vac / (
        1.0 + 2.735182e-4 + 131.4182 / wave_vac**2 + 2.76249e8 / wave_vac**4
    )

    return wave_air


def _load_solar_spectrum():
    """
    Loads the references solar spectrum

    Parameters
    ----------
    None

    Returns
    -------
    array
        Vector with wavelength in air system and flux

    """

    hdu = fits.open(solar_ref_spec)
    tab = hdu[1].data

    wave_air = Quantity(vacuum2air(tab["WAVELENGTH"]), unit=u.AA)
    flux = Quantity(tab["FLUX"], unit=u.erg / (u.cm**2 * u.s * u.AA))

    return wave_air, flux


def sun_magnitude(filters: list[Filter] = [], zeropoint="AB") -> Magnitude:
    """
    Computes the magnitude of Sun in the desired filters

    Parameters
    ----------
    filters: list[Filter]
        Filters as provided by :meth:`milespy.filter.get`
    zeropoint:
        Type of zero point. Valid inputs are AB/VEGA

    Returns
    -------
    Magnitude
        Dictionary with solar mags for each filter

    """
    logger.info("Computing solar absolute magnitudes...")

    wave, flux = _load_solar_spectrum()
    outmags = compute_mags(wave, flux, filters, zeropoint, sun=True)

    return outmags



# -*- coding: utf-8 -*-
import logging
import warnings
from typing import List
from typing import Optional

import h5py
import numpy as np
from astropy import units as u
from astropy.io import ascii
from astropy.units import Quantity
from scipy.spatial import Delaunay
from tqdm import tqdm

from .configuration import get_config_file
from .misc import interp_weights
from .repository import Repository
from .sfh import SFH
from .spectra import Spectra

logger = logging.getLogger("milespy.ssp")

_ssp_attr_units = {
    "age": u.Gyr,
    "alpha": u.dex,
    "met": u.dex,
    "imf_slope": u.dimensionless_unscaled,
}


class SSPLibrary(Repository):
    """
    Single stellar population (SSP) model library.

    This class is used to generate SSP spectra from the repository files.

    Attributes
    ----------
    models: Spectra
        Spectra of all the SSP that form the loaded library
    avail_alphas: list
        Available alpha/Fe values in the loaded library
    avail_imfs: list
        Available initial mass function (IMF) slopes in the loaded library
    source: str
        Name of input models being used
    version: str
        Version number of the models
    """

    warnings.filterwarnings("ignore")

    emiles_lsf = get_config_file("EMILES.lsf")

    # -----------------------------------------------------------------------------
    def __init__(
        self,
        source="MILES_SSP",
        version="9.1",
        isochrone="P",
        imf_type="ch",
    ):
        """
        Creates an instance of the class

        Parameters
        ----------
        source: str, default: "MILES_SSP"
            Name of input models to use.
            If none of the options below, it will assume that `source` is the path
            to an external model repository file.
            Valid inputs are
            MILES_SSP/CaT_SSP/EMILES_SSP
        version: str, default: "9.1"
            version number of the models
        isochrone: str, default: "P"
            Type of isochrone to use. Valid inputs are P/T for Padova+00
            and BaSTI isochrones respectively
        imf_type: str, default: "ch"
            Type of IMF shape. Valid inputs are ch/ku/kb/un/bi

        Notes
        -----
        We limit the choice of models to a given isochrone and imf_type for
        effective loading. Otherwise it can take along time to upload the entire
        dataset

        Returns
        -------
        SSPLibrary

        """
        repo_filename = self._get_repository(source, version)
        self._assert_repository_file(repo_filename)

        # Opening the relevant file in the repository and selecting the desired
        # models at init
        f = h5py.File(repo_filename, "r")
        total_nspec = len(f["age"])

        idx = np.logical_and(
            np.equal(f["imf_type"][...], imf_type.encode()),
            np.equal(f["isochrone"][...], isochrone.encode()),
        )
        self.nspec = np.sum(idx)
        if self.nspec == 0:
            raise ValueError("No cases found with those specs.")

        avail_alphas = np.unique(f["alpha"][idx])
        self.fixed_alpha = len(avail_alphas) == 1
        self.avail_alphas = avail_alphas[~np.isnan(avail_alphas)]

        avail_imfs = np.unique(f["imf_slope"][idx])
        self.avail_imfs = avail_imfs[~np.isnan(avail_imfs)]

        logger.debug(f"{self.nspec} / {total_nspec} cases found")
        logger.debug(f"Fixed_alpha: {self.fixed_alpha}")

        if self.fixed_alpha:
            logger.debug(f"Available alphas: {self.avail_alphas}")

        wave = np.array(f["wave"])
        spec = np.array(f["spec/" + isochrone + "/" + imf_type + "/data"])[...]

        self.source = source
        self.version = version

        fullpath = np.array(np.array(f["filename"][idx]), dtype="str")[0]
        filename = []
        for fullpath in np.array(np.array(f["filename"][idx]), dtype="str"):
            filename.append((fullpath.split("/"))[-1].split(".fits")[0])

        meta = {
            "index": np.arange(self.nspec),
            "isochrone": np.array(np.array(f["isochrone"][idx]), dtype="str"),
            "imf_type": np.array(np.array(f["imf_type"][idx]), dtype="str"),
            "filename": np.array(filename),
        }
        # Standard set of keys available in the repository that are taken
        # care of manually
        base_keys = ["spec", "wave", "isochrone", "imf_type", "filename"]
        # Other information in the repository, store them as arrays
        for k in f.keys():
            if k not in base_keys:
                meta[k] = np.array(f[k])[idx]
                if k in _ssp_attr_units:
                    meta[k] <<= _ssp_attr_units[k]

        f.close()

        super().__init__(
            Spectra(
                spectral_axis=Quantity(wave, unit=u.AA),
                flux=Quantity(spec.T, unit=u.L_sun / u.M_sun / u.AA),
                meta=meta,
            )
        )

        lsf_wave, lsf_fhwm = self._compute_lsf(source)
        self._models.meta["lsf_wave"] = lsf_wave
        self._models.meta["lsf_fwhm"] = lsf_fhwm

        logger.info(source + " models loaded")

    @u.quantity_input
    def in_range(
        self,
        age_lims: u.Quantity[u.Gyr],
        met_lims: u.Quantity[u.dex],
        alpha_lims: Optional[u.Quantity[u.dex]] = None,
        imf_slope_lims: Optional[u.Quantity[u.dimensionless_unscaled]] = None,
        mass: u.Quantity[u.Msun] = u.Quantity(1, unit=u.Msun),
    ) -> Spectra:
        """
        Extracts all SSP models within selected limits

        Parameters
        ----------
        age_lims:
            tuple with age limits
        met_lims:
            tuple with metallicity limits
        alpha_lims:
            tuple with alpha limits
        imf_slope_lims:
            tuple with IMF slope limits
        mass:
            mass of each SSP

        Raises
        ------
        ValueError
            If there is no matching SSP.

        Returns
        -------
        Spectra
            Spectra in the selected ranges

        """

        logger.debug("Searching for models within parameters range")

        if imf_slope_lims is None:
            imf_slope_lims = np.array([0.0, 5.0])

        if self.fixed_alpha or alpha_lims is None:
            idx = (
                (self.models.meta["age"] >= age_lims[0])
                & (self.models.meta["age"] <= age_lims[1])
                & (self.models.meta["met"] >= met_lims[0])
                & (self.models.meta["met"] <= met_lims[1])
                & (self.models.meta["imf_slope"] >= imf_slope_lims[0])
                & (self.models.meta["imf_slope"] <= imf_slope_lims[1])
            )
        else:
            idx = (
                (self.models.meta["age"] >= age_lims[0])
                & (self.models.meta["age"] <= age_lims[1])
                & (self.models.meta["met"] >= met_lims[0])
                & (self.models.meta["met"] <= met_lims[1])
                & (self.models.meta["alpha"] >= alpha_lims[0])
                & (self.models.meta["alpha"] <= alpha_lims[1])
                & (self.models.meta["imf_slope"] >= imf_slope_lims[0])
                & (self.models.meta["imf_slope"] <= imf_slope_lims[1])
            )

        ncases = np.sum(idx)
        logger.debug(f"{ncases} cases found")

        if ncases == 0:
            raise ValueError("No matching SSPs")

        out = Spectra.__getitem__(self.models, idx)._apply_mass(mass)

        return out

    @u.quantity_input
    def in_list(
        self,
        age: u.Quantity[u.Gyr],
        met: u.Quantity[u.dex],
        alpha: Optional[u.Quantity[u.dex]] = None,
        imf_slope: Optional[u.Quantity[u.dimensionless_unscaled]] = None,
        mass: Optional[u.Quantity[u.Msun]] = None,
    ) -> Spectra:
        """
        Extracts a selected set of models available from the library.

        Parameters
        ----------
        age :
            list of ages to extract
        met :
            list of metallicities to extract
        alpha :
            list of alphas to extract
        imf_slope :
            list of IMF slopes to extract
        mass : default 1 solar mass
            mass of each SSP

        Notes
        -----
        All lists must have the same length.
        This function does not perform any interpolation.
        Values in the inputs have to be valid ages, mets, alpha, and imf_slopes
        for the input isochrone and imf_type.

        Raises
        ------
        ValueError
            If inputs do not have the same shape or there is no resulting SSP.

        Warns
        -----
        If the number of output spectra is different that the input values.

        Returns
        -------
        Spectra

        """
        if alpha is not None and self.fixed_alpha:
            raise ValueError(
                "This repository does not provide variable alpha:\n"
                + "Source: "
                + self.source
                + "Isochrone: "
                + self.models.meta["isochrone"][0]
                + "IMF: "
                + self.models.meta["imf_type"][0]
            )

        # Treating the input list
        age = np.array(age).ravel()
        met = np.array(met).ravel()
        alpha = np.array(alpha).ravel()
        imf_slope = np.array(imf_slope).ravel()

        # Checking they have the same number of elements
        if self.fixed_alpha or alpha is None:
            if len(set(map(len, (age, met, imf_slope)))) != 1:
                raise ValueError("Input list values do not have the same length.")
        else:
            if len(set(map(len, (age, met, alpha, imf_slope)))) != 1:
                raise ValueError("Input list values do not have the same length.")

        ncases = len(age)
        logger.debug("Searching for {ncases} SSPs")

        if mass is None:
            mass = Quantity(value=np.ones(ncases), unit=u.Msun)

        id = []
        for i in range(ncases):
            if self.fixed_alpha or alpha is None:
                idx = (
                    (np.array(self.models.meta["age"]) == age[i])
                    & (np.array(self.models.meta["met"]) == met[i])
                    & (np.array(self.models.meta["imf_slope"]) == imf_slope[i])
                )
            else:
                idx = (
                    (np.array(self.models.meta["age"]) == age[i])
                    & (np.array(self.models.meta["met"]) == met[i])
                    & (np.array(self.models.meta["alpha"]) == alpha[i])
                    & (np.array(self.models.meta["imf_slope"]) == imf_slope[i])
                )

            if np.any(idx):
                id = np.append(id, self.models.meta["index"][idx])

        good = np.array(id, dtype=int)
        ngood = len(good)
        logger.debug(f"{ngood} cases found")

        if ncases != ngood:
            if ngood == 0:
                raise ValueError(
                    "Check those values exist for isochrone: "
                    + self.models.meta["isochrone"][0]
                    + " and IMF_Type: "
                    + self.models.meta["imf_type"][0]
                    + ". Please check this agrees with init instance."
                )
            else:
                logger.warning(
                    f"Asked for {ncases} SSPs, but found only {ngood} matching ones"
                )

        out = Spectra.__getitem__(self.models, good)._apply_mass(mass)

        return out

    @u.quantity_input
    def closest(
        self,
        age: u.Quantity[u.Gyr],
        met: u.Quantity[u.dex],
        alpha: Optional[u.Quantity[u.dex]] = None,
        imf_slope: Optional[u.Quantity[u.dimensionless_unscaled]] = None,
        mass: u.Quantity[u.Msun] = Quantity(value=1.0, unit=u.Msun),
    ) -> Spectra:
        """
        Retrieve the closest SSP avaiable in the library

        Parameters
        ----------
        age:
            Desired age
        met:
            Desired metallicity
        alpha:
            Desired alpha
        imf_slope:
            Desired IMF slope
        mass:
            Mass of the SSP

        Returns
        -------
        Spectra
            Closest spectra from the repository.

        Raises
        ------
        RuntimeError
            If the values are out of the grid.
        ValueError
            If the provided parameters do not have the same shape.
        """
        return self.interpolate(
            age=age, met=met, alpha=alpha, imf_slope=imf_slope, mass=mass, closest=True
        )

    @u.quantity_input
    def interpolate(
        self,
        age: u.Quantity[u.Gyr],
        met: u.Quantity[u.dex],
        alpha: Optional[u.Quantity[u.dex]] = None,
        imf_slope: Optional[u.Quantity[u.dimensionless_unscaled]] = None,
        mass: u.Quantity[u.Msun] = Quantity(value=1.0, unit=u.Msun),
        closest: bool = False,
        simplex: bool = False,
        force_interp: List = [],
    ) -> Spectra:
        """
        Interpolates SSP models for certain parameters using Delaunay triangulation

        Parameters
        ----------
        age:
            Desired age
        met:
            Desired metallicity
        alpha:
            Desired alpha
        imf_slope:
            Desired IMF slope
        mass:
            Mass of the SSP
        closest:
            Return the closest spectra, rather than performing the interpolation.
            If only one interpolation is performed, all the spectra in the simplex
            vertices are returned.
        simplex:
            If only one set of input parameters is given, return all the spectra
            that form part of the simplex used for the interpolation. These spectra
            have the weights information in their `meta` dictionary.
        force_interp:
            Force the interpolation over the indicated variables, even if the
            asked alpha/imf_slope is sampled in the repository. Valid values
            are "alpha" and "imf_slope".

        Returns
        -------
        Spectra
            Interpolated spectrum.
            If closest == True, return the closest spectra from the repository,
            rather than doing the interpolation.

        Raises
        ------
        RuntimeError
            If the values are out of the grid.
        ValueError
            If the provided parameters do not have the same shape.
        """
        # Preprocess the input
        single_alpha = alpha is not None and np.ndim(alpha) == 0
        nan_alpha = alpha is None
        single_imf_slope = imf_slope is not None and np.ndim(imf_slope) == 0
        nan_imf = imf_slope is None

        age = np.atleast_1d(age)
        met = np.atleast_1d(met)
        alpha = np.atleast_1d(alpha)
        # For the IMF slope we can have floats being passed, so we make sure
        # that it is a Quantity object
        imf_slope = np.atleast_1d(imf_slope)
        if not nan_imf:
            imf_slope <<= u.dimensionless_unscaled

        wrong_shape = age.shape != met.shape
        if not nan_alpha:
            wrong_shape |= age.shape != alpha.shape
        if not nan_imf:
            wrong_shape |= age.shape != imf_slope.shape
        if wrong_shape:
            raise ValueError("The input parameters should all have the same shape")

        ninterp = len(age)

        # Checking input point is within the grid
        def _compute_bounds(val, model):
            min_model = np.amin(model)
            max_model = np.amax(model)
            if np.any(val < min_model) or np.any(val > max_model):
                raise RuntimeError(
                    f"Input parameters {val} is outside of model grid: "
                    f"({min_model},{max_model})"
                )

        _compute_bounds(age, self.models.meta["age"])
        _compute_bounds(met, self.models.meta["met"])

        if not nan_imf:
            _compute_bounds(imf_slope, self.models.meta["imf_slope"])

        if not self.fixed_alpha and not nan_alpha:
            _compute_bounds(alpha, self.models.meta["alpha"])

        ndims = 2
        extra_dim = 0

        if nan_imf and len(self.avail_imfs) > 1:
            raise ValueError(f"An IMF slope should be provided from: {self.avail_imfs}")

        # If there is a single available imf slope, there is no need to interpolate
        # over it. Also, if the users gives an imf_slope that **exactly** matches
        # one of the database, we fix the imf_slope for the interpolation
        interp_fix_imf_slope = len(self.avail_imfs) == 1 or (
            single_imf_slope and imf_slope in self.avail_imfs
        )
        # But this can be overruled
        interp_fix_imf_slope &= not ("imf_slope" in force_interp)
        if interp_fix_imf_slope:
            if nan_imf:
                imf_slope = self.avail_imfs[0]

            imf_mask = np.equal(self.models.meta["imf_slope"], imf_slope[0])
        else:
            imf_mask = np.full(self.models.meta["imf_slope"].shape, True)
            ndims += 1

        logger.debug(f"Fixed imf during the interpolation? {interp_fix_imf_slope}")

        # Same as above for the imf_slope, however, now we can have nan values in
        # the repository in the case of "base" models.
        interp_fix_alpha = (
            self.fixed_alpha
            or nan_alpha
            or (single_alpha and alpha in self.avail_alphas)
        )
        if self.fixed_alpha and not nan_alpha:
            logger.warning("There is no alpha-enhanced SSPs with this model choice")
        interp_fix_alpha &= not ("alpha" in force_interp)
        if interp_fix_alpha:
            if nan_alpha:
                alpha_mask = np.isnan(self.models.meta["alpha"])
            elif alpha[0] in self.avail_alphas:
                alpha_mask = np.equal(self.models.meta["alpha"], alpha[0])
            else:
                # There is a single alpha in all the repository, so no need to mask
                alpha_mask = np.full(self.models.meta["alpha"].shape, True)
        else:
            # Remove the base alpha (i.e., nans) from the interpolation
            alpha_mask = ~np.isnan(self.models.meta["alpha"])
            ndims += 1

        logger.debug(f"Fixed alpha during the interpolation? {interp_fix_alpha}")

        idx = alpha_mask & imf_mask
        # This is required for future extrapolation, e.g., in the SFH module
        self.idx = idx

        n_avail = np.sum(idx)

        logger.debug(f"Interpolating over a grid of {n_avail} spectra")
        logger.debug(f"Creating {ndims}-D Delaunay triangulation")

        self.params = np.empty((n_avail, ndims))
        self.params[:, 0] = self.models.meta["age"][idx]
        self.params[:, 1] = self.models.meta["met"][idx]

        if not interp_fix_imf_slope:
            self.params[:, 2 + extra_dim] = self.models.meta["imf_slope"][idx]
            extra_dim += 1

        if not interp_fix_alpha:
            self.params[:, 2 + extra_dim] = self.models.meta["alpha"][idx]

        self.tri = Delaunay(self.params, qhull_options="QJ")

        if closest:
            closest_idx = np.empty(ninterp, dtype=int)
        else:
            wave = self.models.spectral_axis
            spec = Quantity(
                value=np.empty((ninterp, self.models.data.shape[1])),
                unit=self.models.flux.unit,
            )

            new_meta = {
                "imf_type": np.full(ninterp, self.models.meta["imf_type"][0]),
                "met": met,
                "age": age,
                "lsf_wave": self.models.meta["lsf_wave"],
                "lsf_fwhm": self.models.meta["lsf_fwhm"],
            }
            if interp_fix_alpha:
                new_meta["alpha"] = np.full(ninterp, alpha)
                if not nan_alpha:
                    new_meta["alpha"] <<= u.dex
            if interp_fix_imf_slope:
                new_meta["imf_slope"] = np.full(ninterp, imf_slope)

            base_keys = list(new_meta.keys())
            for k in self.models.meta.keys():
                if k not in new_meta.keys():
                    if len(self.models.meta[k]) > 1:
                        if "U" not in self.models.meta[k].dtype.kind:
                            new_meta[k] = np.empty(
                                ninterp, dtype=self.models.meta[k].dtype
                            )
                            if hasattr(self.models.meta[k], "unit"):
                                new_meta[k] <<= self.models.meta[k].unit

        for i in tqdm(range(ninterp), delay=3.0):
            input_pt = [age[i].to_value(u.Gyr), met[i].to_value(u.dex)]
            if not interp_fix_imf_slope:
                input_pt.append(imf_slope[i].value)
            if not interp_fix_alpha:
                input_pt.append(alpha[i].value)

            vtx, wts = interp_weights(self.params, np.atleast_2d(input_pt), self.tri)
            vtx, wts = vtx.ravel(), wts.ravel()

            if logger.isEnabledFor(logging.DEBUG):
                logger.debug(f"Searching for the simplex around: {input_pt}")
                logger.debug(f"Simplex ids: {self.models.meta['index'][idx][vtx]}")
                logger.debug(f"Simple ages: {self.models.meta['age'][idx][vtx]}")
                logger.debug(f"Simplex met: {self.models.meta['met'][idx][vtx]}")
                logger.debug(f"Simplex weights: {wts}, norm: {np.sum(wts)}")

            if closest:
                if simplex and ninterp == 1:
                    out = Spectra.__getitem__(
                        self.models, self.models.meta["index"][idx][vtx]
                    )._apply_mass(mass)
                    return out
                else:
                    closest_idx[i] = vtx[np.argmax(wts)]

            else:
                spec[i, :] = np.dot(self.models.flux[idx, :][vtx].T, wts)

                # Interpolate the rest of the meta if possible
                for k in new_meta:
                    if len(self.models.meta[k]) > 1:
                        if k not in base_keys:
                            new_meta[k][i] = np.dot(self.models.meta[k][idx][vtx], wts)

        if closest:
            out = Spectra.__getitem__(self.models, closest_idx)._apply_mass(mass)
        else:
            out = Spectra(spectral_axis=wave, flux=spec, meta=new_meta)._apply_mass(
                mass
            )

        # Select the first and only spectra so that the users does not need to
        # do this all the time, manually
        if ninterp == 1:
            out = out[0]

        return out

    def _compute_lsf(self, source: str):
        # This information could be given in the repository files!
        """
        Returns the line-spread function (LSF) given a source

        Parameters
        ----------
        source:
            Name of input models to use. Valid inputs are MILES_SSP/CaT_SSP/EMILES_SSP


        Returns
        -------
        lsf_wave: Quantity
            Wavelenghts of the LSF
        lsf_fwhm: Quantity
            Value of the LSF in terms of full-width half-maximum
        """

        lsf_wave = self.models.spectral_axis
        npix = self.models.npix
        if source == "MILES_SSP":
            lsf_fwhm = 2.51 * np.ones(npix)

        elif source == "MILES_STARS":
            lsf_fwhm = 2.50 * np.ones(npix)

        elif source == "CaT_SSP":
            lsf_fwhm = 1.50 * np.ones(npix)

        elif source == "CaT_STARS":
            lsf_fwhm = 1.50 * np.ones(npix)

        elif source == "EMILES_SSP":
            tab = ascii.read(self.emiles_lsf)
            wave = tab["col1"]
            fwhm = tab["col2"]
            lsf_fwhm = np.interp(lsf_wave.to_value(u.AA), wave, fwhm)

        else:
            raise ValueError(
                self.source
                + " is not a valid entry."
                + "Allowed values: MILES_SSP/MILES_STARS/CaT_SSP/CaT_STARS/EMILES"
            )

        return lsf_wave, lsf_fwhm * u.AA

    def from_sfh(self, sfh: SFH) -> Spectra:
        """
        Compute the spectra derived from the input SFH.

        Parameters
        ----------
        sfh:
            Star formation history

        Returns
        -------
        Spectra
        """

        # We make an initial call to interpolate (ssp_model_class) to
        # obtain the values of the triangulation to be used extensively below
        interp_specs = self.interpolate(
            age=sfh.time,
            met=sfh.met,
            imf_slope=sfh.imf,
            mass=sfh.time_weights,
            alpha=sfh.alpha,
            force_interp=["alpha", "imf_slope"],
        )

        new_meta = {
            "lsf_wave": self.models.meta["lsf_wave"],
            "lsf_fwhm": self.models.meta["lsf_fwhm"],
        }
        for k in self.models.meta.keys():
            if len(self.models.meta[k]) > 1:
                # Skip the interpolation of string data, e.g., filenames
                kind = self.models.meta[k].dtype.kind
                if "f" in kind:
                    new_meta[k] = np.zeros(1)
                    if hasattr(self.models.meta[k], "unit"):
                        new_meta[k] <<= self.models.meta[k].unit

        # We iterate now over all the age bins in the SFH
        for t, date in tqdm(enumerate(sfh.time), delay=3.0):
            # Update quantities
            for k in new_meta.keys():
                if len(interp_specs.meta[k]) == interp_specs.nspec:
                    new_meta[k] += interp_specs.meta[k][t] * (
                        sfh.time_weights[t] / sfh.mass
                    )

        # The final spectrum is also the mass-weighted one, but the mass was
        # already taking into account when doing the interpolation
        ospec = np.sum(interp_specs.flux, axis=0)

        return Spectra(
            spectral_axis=self.models.spectral_axis, flux=ospec, meta=new_meta
        )



# -*- coding: utf-8 -*-
import logging
from typing import Optional

import numpy as np
from astropy import units as u
from scipy.integrate import trapezoid

logger = logging.getLogger("milespy.sfh")

DEFAULT_NBINS = 20


class SFH:
    """
    Class for manipulating star formation histories (SFH) and create derived spectra

    Attributes
    ----------
    time: ~astropy.units.Quantity
        Look-back time
    sfr: ~astropy.units.Quantity
        Values of the star formation rate (SFR) at each time
    met: ~astropy.units.Quantity
        Values of the metallicity at each time (dex)
    alpha: ~astropy.units.Quantity
        Values of [alpha/Fe] at each time (dex)
    imf: ~astropy.units.Quantity
        Values of the IMF slope
    """

    @u.quantity_input
    def __init__(
        self, time: u.Quantity[u.Gyr] = np.linspace(0.035, 13.5, DEFAULT_NBINS) << u.Gyr
    ):
        """
        Create a base SFH object

        Parameters
        ----------
        time: ~astropy.units.Quantity
            Look-back time array for the SFH samples
        """
        nbins = len(time)
        self.time = time
        self.sfr = np.zeros(nbins) << u.Msun / u.Gyr
        self.met = np.zeros(nbins) << u.dex
        self.alpha = np.zeros(nbins) << u.dex
        self.imf = np.full(nbins, 1.3) << u.dimensionless_unscaled

    @staticmethod
    def _process_param(argname, arg, refname, ref, offset=0):
        if np.isscalar(arg):
            arg = np.full_like(ref, arg)
        elif len(arg) == len(ref) + offset:
            arg = np.array(arg)
        else:
            raise ValueError(
                f"{refname} and {argname} arrays should be the same length"
            )

    @staticmethod
    def _validate_scalar(arg, argname="Input"):
        if np.ndim(arg) != 0:
            raise ValueError(f"{argname} should be scalar")

    @staticmethod
    def _validate_in_range(arg, low, high, argname="Input"):
        if arg < low or arg > high:
            raise ValueError(f"{argname} is out of range")

    def _compute_time_weights(self):
        dt = self.time - np.roll(self.time, 1)
        dt[0] = dt[1]
        self.time_weights = (self.sfr * dt).to(u.Msun)

    def _normalize(self, sfr):
        norm = trapezoid(sfr, x=self.time.to_value(u.yr))
        self.sfr = sfr / norm * self.mass / u.yr

        self._compute_time_weights()

    @u.quantity_input
    def sfr_tau(
        self,
        start: u.Quantity[u.Gyr] = 10.0 * u.Gyr,
        tau: u.Quantity[u.Gyr] = 1.0 * u.Gyr,
        mass: u.Quantity[u.Msun] = 1.0 * u.Msun,
    ):
        r"""Exponentially declining SFR

        This is a standard tau model where the SFR is given by

        .. math::

            \text{SFR}(t) =
            \begin{cases}
            0 & \text{if}\; t < t_0 \\
            e^{- (t-t_0)/\tau } & \text{if}\; t >= t_0 \\
            \end{cases}


        Parameters
        ----------

        start : ~astropy.units.Quantity
            Start of the burst (default=10 Gyr)
        tau   : ~astropy.units.Quantity
            e-folding time (default=1 Gyr)
        mass : ~astropy.units.Quantity
            Total formed mass (default=1 Msun)
        """
        for inp in (start, tau, mass):
            self._validate_scalar(inp)

        # Exponentially declining SFR
        self.time <= start
        sfr = np.zeros(self.time.shape)
        sfr[(self.time <= start)] = np.exp(
            -(start - self.time[(self.time <= start)]) / tau
        )
        self.mass = mass
        self._normalize(sfr)

    @u.quantity_input
    def sfr_delayed_tau(
        self,
        start: u.Quantity[u.Gyr] = 10.0 * u.Gyr,
        tau: u.Quantity[u.Gyr] = 1.0 * u.Gyr,
        mass: u.Quantity[u.Msun] = 1.0 * u.Msun,
    ):
        r"""Delayed exponentially declining SFR

        This is a standard tau model where the SFR is given by

        .. math::

            \text{SFR}(t) =
            \begin{cases}
            0 & \text{if}\; t < t_0 \\
            (t_0-t)e^{- (t-t_0)/\tau } & \text{if}\; t >= t_0 \\
            \end{cases}

        Parameters
        ----------

        start : ~astropy.units.Quantity
            Start of the burst (default=10 Gyr)
        tau   : ~astropy.units.Quantity
            e-folding time (default=1 Gyr)
        mass : ~astropy.units.Quantity
            Total formed mass (default=1 Msun)

        """

        for inp in (start, tau):
            self._validate_scalar(inp)

        sfr = np.zeros(self.time.shape)
        sfr[(self.time <= start)] = (start - self.time[(self.time <= start)]) * np.exp(
            -(start - self.time[(self.time <= start)]) / tau
        )

        self.mass = mass
        self._normalize(sfr)

    @u.quantity_input
    def sfr_lognormal(
        self,
        Tpeak: u.Quantity[u.Gyr] = 10.0 * u.Gyr,
        tau: u.Quantity[u.Gyr] = 1.0 * u.Gyr,
        mass: u.Quantity[u.Msun] = 1.0 * u.Msun,
    ):
        r"""Lognormal SFR

        The time evolution of the SFR is given by

        .. math::

            \text{SFR}(t) = e^{ -(t_0- \log(t_n))^2 / 2\tau^2} / t_n


        Note that :math:`t_n` is in this case the time since the Big Bang and not
        lookback time as in the SSP. See details in Diemer et al. 2017,
        ApJ, 839, 26, Appendix A.1


        Parameters
        ----------

        Tpeak : ~astropy.units.Quantity
            Time of the SFR peak (default=10 Gyr)
        tau   : ~astropy.units.Quantity
            Characteristic time-scale (default=1 Gyr)
        mass : ~astropy.units.Quantity
            Total formed mass (default=1 Msun)
        """

        for inp in (Tpeak, tau):
            self._validate_scalar(inp)

        time = self.time.to_value(u.Gyr)
        tau = tau.to_value(u.Gyr)
        Tpeak = Tpeak.to_value(u.Gyr)
        self.mass = mass

        # While SSPs are based on lookback times we will use in this case
        # time since the oldest age in the grid (~ time since the Big Bang)
        tn = np.max(time) - time
        tn[-1] = 1e-4  # Avoid zeros in time...
        Tc = np.log(time.max() - Tpeak) + tau**2
        sfr = np.zeros(self.time.shape)
        sfr = (1 / tn) * np.exp(-((np.log(tn) - Tc) ** 2) / (2.0 * tau**2))

        self._normalize(sfr)

    @u.quantity_input
    def sfr_double_power_law(
        self,
        a=5,
        b=5,
        tp: u.Quantity[u.Gyr] = 10 * u.Gyr,
        mass: u.Quantity[u.Msun] = 1.0 * u.Msun,
    ):
        r"""Double power law SFR evolution

        The SFR as a function of time is given by (Behroozi et al. 2013)

        .. math::

            \text{SFR}(t_n) = ((t_n/t_p)^a + (t_n/t_p)^{-b})^{-1}

        As for the lognormal SFR, :math:`t_n` refers to time since the Big Bang


        Parameters
        ----------
        a     : float
            falling slope (default=5)
        b     : float
            rising slope (default=5)
        tp   : ~astropy.units.Quantity
            similar to the SFR peak in look-back time (default=10 Gyr)
        mass : ~astropy.units.Quantity
            Total formed mass (default=1 Msun)
        """

        for inp in (a, b, tp):
            self._validate_scalar(inp)

        self.mass = mass

        # While SSPs are based on lookback times we will use in this case
        # time since the oldest age in the grid (~ time since the Big Bang)
        tn = max(self.time) - self.time
        tau = max(self.time) - tp
        tn[-1] = 1e-4 * u.Gyr  # Avoid zeros in time...
        sfr = np.zeros_like(self.time)
        sfr = ((tn / tau) ** a + (tn / tau) ** -b) ** -1

        self._normalize(sfr)

    @u.quantity_input
    def sfr_custom(
        self, sfr: u.Quantity[u.Msun / u.yr], mass: Optional[u.Quantity[u.Msun]] = None
    ):
        """User-defined SFR

        Parameters
        ----------
        sfr     : ~astropy.units.Quantity
            Star formation rate array, with the same shape as `time`.
        mass     : ~astropy.units.Quantity
            If given, normalize the input SFR such that it produces this mass.

        """
        if mass is None:
            self._set_input_sfr(sfr)
        else:
            self.mass = mass
            self._normalize(sfr)

    @u.quantity_input
    def _set_input_sfr(self, sfr: u.Quantity[u.Msun / u.Gyr]):
        self.mass = trapezoid(sfr, x=self.time)
        self.sfr = sfr
        self._compute_time_weights()

    @staticmethod
    def _linear(time, start, end, t_start, t_end):
        for inp in (start, end, t_start, t_end):
            SFH._validate_scalar(inp)

        slope = (start - end) / (t_start - t_end)
        out = np.empty(time.shape)

        m = time > t_start
        out[m] = start

        m = (time <= t_start) & (time >= t_end)
        out[m] = slope * (time[m] - t_start) + start

        m = time < t_end
        out[m] = end

        return out

    @staticmethod
    def _sigmoid(time, start, end, tc, gamma):
        for inp in (start, end, tc, gamma):
            SFH._validate_scalar(inp)

        return (end - start) / (1.0 + np.exp(-gamma * (tc - time))) + start

    @u.quantity_input
    def met_sigmoid(
        self,
        start: u.Quantity[u.dex] = -1.5,
        end: u.Quantity[u.dex] = 0.2,
        tc: u.Quantity[u.Gyr] = 5.0 * u.Gyr,
        gamma: u.Quantity[1 / u.Gyr] = 1.0 / u.Gyr,
    ):
        """Sigmoidal metallicity evolution

        The metallicity evolves as a sigmoidal function. This is not
        meant to be physically meaningful but to reproduce the exponential
        character of the chemical evolution

        Parameters
        ----------

        start   : ~astropy.units.Quantity (dex)
            Metallicity of the oldest stellar population (default=-1.5 dex)
        end     : ~astropy.units.Quantity (dex)
            Metallicity of the youngest stellar population (default=0.2 dex)
        tc      : ~astropy.units.Quantity (Gyr)
            Characteristic transition time (default=5 Gyr)
        gamma   : ~astropy.units.Quantity
            Transition slope (default=1/Gyr)

        """
        self.met = self._sigmoid(self.time, start, end, tc, gamma)

    @u.quantity_input
    def alpha_sigmoid(
        self,
        start: u.Quantity[u.dex] = 0.4,
        end: u.Quantity[u.dex] = 0.0,
        tc: u.Quantity[u.Gyr] = 5.0 * u.Gyr,
        gamma: u.Quantity[1 / u.Gyr] = 1.0 / u.Gyr,
    ):
        """Sigmoidal [alpha/Fe] evolution

        The [alpha/Fe] evolves as a sigmoidal function. This is not
        meant to be physically meaningful but to reproduce the exponential
        character of the chemical evolution

        Parameters
        ----------

        start   : ~astropy.units.Quantity (dex)
            [alpha/Fe] of the oldest stellar population (default=-1.5 dex)
        end     : ~astropy.units.Quantity (dex)
            [alpha/Fe] of the youngest stellar population (default=0.2 dex)
        tc      : ~astropy.units.Quantity (Gyr)
            Characteristic transition time (default=5 Gyr)
        gamma   : ~astropy.units.Quantity
            Transition slope (default=1/Gyr)

        """
        self.alpha = self._sigmoid(self.time, start, end, tc, gamma)

    @u.quantity_input
    def met_linear(
        self,
        start: u.Quantity[u.dex] = -1.5,
        end: u.Quantity[u.dex] = 0.2,
        t_start: u.Quantity[u.Gyr] = 5.0 * u.Gyr,
        t_end: u.Quantity[u.Gyr] = 1.0 * u.Gyr,
    ):
        """Linear metallicity evolution

        The metallicity evolves as a ReLU function, i.e., constant at the beginning
        and linearly varing afterwards.

        Parameters
        ----------

        start   : ~astropy.units.Quantity (dex)
            Metallicity of the oldest stellar population (default=-1.5 dex)
        end     : ~astropy.units.Quantity (dex)
            Metallicity of the youngest stellar population (default=0.2 dex)
        t_start : ~astropy.units.Quantity (Gyr)
            Start of the metallicity variation (default=5 Gyr)
        t_end : ~astropy.units.Quantity (Gyr)
            End of the metallicity variation (default=5 Gyr)
        """
        self.met = self._linear(self.time, start, end, t_start, t_end)

    @u.quantity_input
    def alpha_linear(
        self,
        start: u.Quantity[u.dex] = -1.5,
        end: u.Quantity[u.dex] = 0.2,
        t_start: u.Quantity[u.Gyr] = 5.0 * u.Gyr,
        t_end: u.Quantity[u.Gyr] = 1.0 * u.Gyr,
    ):
        """Linear [alpha/Fe] evolution

        The [alpha/Fe] evolves as a ReLU function, i.e., constant at the beginning
        and linearly varing afterwards.

        Parameters
        ----------

        start   : ~astropy.units.Quantity (dex)
            [alpha/Fe] of the oldest stellar population (default=-1.5 dex)
        end     : ~astropy.units.Quantity (dex)
            [alpha/Fe] of the youngest stellar population (default=0.2 dex)
        t_start : ~astropy.units.Quantity (Gyr)
            Start of the [alpha/Fe] variation (default=5 Gyr)
        t_end : ~astropy.units.Quantity (Gyr)
            End of the [alpha/Fe] variation (default=1 Gyr)
        """
        self.alpha = self._linear(self.time, start, end, t_start, t_end)

    @u.quantity_input
    def imf_linear(
        self,
        start: u.Quantity = -1.5,
        end: u.Quantity = 0.2,
        t_start: u.Quantity[u.Gyr] = 5.0 * u.Gyr,
        t_end: u.Quantity[u.Gyr] = 1.0 * u.Gyr,
    ):
        """Linear IMF slope evolution

        The IMF slope evolves as a ReLU function, i.e., constant at the beginning
        and linearly varing afterwards.

        Parameters
        ----------

        start   : ~astropy.units.Quantity
            IMF of the oldest stellar population (default=-1.5)
        end     : ~astropy.units.Quantity
            IMF  of the youngest stellar population (default=0.2)
        t_start : ~astropy.units.Quantity (Gyr)
            Start of the IMF variation (default=5 Gyr)
        t_end : ~astropy.units.Quantity (Gyr)
            End of the IMF variation (default=1 Gyr)
        """
        self.imf = self._linear(self.time, start, end, t_start, t_end)

    @u.quantity_input
    def imf_sigmoid(
        self,
        start: u.Quantity = 0.5,
        end: u.Quantity = 3.0,
        tc: u.Quantity[u.Gyr] = 5.0 * u.Gyr,
        gamma: u.Quantity[1 / u.Gyr] = 1.0 / u.Gyr,
    ):
        """Sigmoidal IMF slope evolution

        The IMF slope  evolves as a sigmoidal function. This is not
        meant to be physically meaningful but to track the chemical
        variations (see e.g. Martin-Navarro et al. 2015)

        Parameters
        ----------

        start   : ~astropy.units.Quantity (dex)
            IMF slope  of the oldest stellar population (default=0.5)
        end     : ~astropy.units.Quantity (dex)
            IMF slope  of the youngest stellar population (default=3.0)
        tc      : ~astropy.units.Quantity (Gyr)
            Characteristic transition time (default=5 Gyr)
        gamma   : ~astropy.units.Quantity
            Transition slope (default=1 Gyr)

        """
        self.imf = self._sigmoid(self.time, start, end, tc, gamma)



# -*- coding: utf-8 -*-
import logging
import pathlib

from rcfile import rcfile


def _setup_logger(config):
    logger = logging.getLogger("milespy")
    logger.setLevel(logging.WARNING)

    for existing_handler in list(logger.handlers):
        logger.removeHandler(existing_handler)

    ch = logging.StreamHandler()
    ch.setFormatter(logging.Formatter("%(name)s: %(message)s"))
    logger.addHandler(ch)

    if "log" in config.keys():
        logger.setLevel(config["log"])


config = rcfile("milespy", {})
config_folder = pathlib.Path(__file__).parent.resolve() / "config_files"
def_repo_folder = pathlib.Path(__file__).parent.resolve()

logger = _setup_logger(config)


def get_config_file(name):
    return config_folder.as_posix() + "/" + name



# -*- coding: utf-8 -*-
import numpy as np


def interp_weights(xyz, uvw, tri):
    """
    Creates a Delaunay triangulation and finds the vertices and weights of
    points around a given location in parameter space
    """

    d = len(uvw[0, :])
    simplex = tri.find_simplex(uvw)
    vertices = np.take(tri.simplices, simplex, axis=0)
    temp = np.take(tri.transform, simplex, axis=0)
    delta = uvw - temp[:, d]
    bary = np.einsum("njk,nk->nj", temp[:, :d, :], delta)

    return vertices, np.hstack((bary, 1 - bary.sum(axis=1, keepdims=True)))



# -*- coding: utf-8 -*-

__version__ = "1.0rc3"


from .ssp_library import SSPLibrary
from .stellar_library import StellarLibrary
from .sfh import SFH

__all__ = ["SSPLibrary", "StellarLibrary", "SFH"]



# -*- coding: utf-8 -*-
from __future__ import annotations

import logging
import typing
import warnings
from copy import copy

import numpy as np
from astropy import units as u
from astropy.constants import c
from spectres import spectres
from specutils import Spectrum1D
from tqdm import tqdm

from .configuration import get_config_file
from .filter import Filter
from .ls_indices import LineStrengthDict
from .ls_indices import LineStrengthIndex
from .ls_indices import lsindex
from .magnitudes import compute_mags
from .magnitudes import Magnitude
from .magnitudes import sun_magnitude

# ==============================================================================

logger = logging.getLogger("milespy.spectra")


class Spectra(Spectrum1D):
    """
    Class for storing spectra in milespy.

    This class inherits `Spectrum1D` from specutils, and can use
    all the methods of that class.

    The main difference is how we use the `meta` dictionary.
    We assume that all the values stored in meta have the same lengths as the
    number of spectra.  Thus, each value of each key univoquely refers to some
    information of a given spectra.  The exact keys in this dictionary will
    depend on the model creating this object.

    """

    warnings.filterwarnings("ignore")

    solar_ref_spec = get_config_file("sun_mod_001.fits")

    @property
    def properties(self):
        return list(self.meta.keys())

    def __getattr__(self, attr):
        if attr not in ["meta", "_meta"]:
            if attr in self.meta.keys():
                if (
                    np.ndim(self.meta[attr]) > 0
                    and hasattr(self.meta[attr], "__len__")
                    and len(self.meta[attr]) == 1
                ):
                    return self.meta[attr][0]
                else:
                    return self.meta[attr]
            else:
                raise AttributeError

    @property
    def npix(self):
        if len(self.data.shape) > 1:
            return self.data.shape[1]
        else:
            return self.data.shape[0]

    @property
    def dim(self):
        return self.data.shape[:-1]

    @property
    def nspec(self):
        if len(self.data.shape) > 1:
            return np.prod(self.data.shape[:-1])
        else:
            return 1

    def __getitem__(self, item):
        out = super().__getitem__(item)
        for k in out.meta.keys():
            try:
                if len(out.meta[k]) == self.nspec:
                    out.meta[k] = out.meta[k][item]
                else:
                    out.meta[k] = self.meta[k]
            except TypeError:
                pass
        return out

    def redshift_spectra(self, redshift=None):
        # This may still be required because it also changes the LSF
        """
        Returns a copy of the instance with a redshifted wavelength vector,
        spectra and LSF

        Parameters
        ----------
        redshift:
            Desired redshift

        Returns
        -------
        Spectra
            Object instance with spectra redshifted and updated info

        """
        logger.info("Redshifting spectra ...")

        out = copy(self)
        # wave = out.wave * (1.0 + redshift)
        # spec = out.spec / (1.0 + redshift)
        # out.update_basic_pars(wave, spec)
        out.redshift = redshift
        out.lsf_fwhm = out.lsf_fwhm / (1.0 + redshift)

        return out

    # @u.quantity_input # For some reason this does not work...
    def velocity_shift(self, v_los: u.Quantity[u.km / u.s]) -> Spectra:
        if v_los.isscalar:
            # Simply move the spectral axis
            return Spectra(
                spectral_axis=self.spectral_axis * (1.0 + v_los / c),
                flux=self.flux,
                meta=copy(self.meta),
            )
        else:
            if len(v_los) != self.nspec:
                raise ValueError(
                    "The size of the input velocity shift should"
                    "match the number of spectra"
                )

            # Preallocate for performance
            new_flux = np.empty(self.flux.shape)
            ref_axis = self.spectral_axis.to_value(u.AA)
            new_axis = np.empty_like(self.spectral_axis.to_value(u.AA))

            for i, vi in tqdm(enumerate(v_los), delay=3.0, total=self.nspec):
                # NB: This would be ideal to conserve flux, but it is very
                #   computationally expensive.
                # new_flux[i, :] = spectres(
                #     self.spectral_axis.to_value(u.AA),
                #     self.spectral_axis.to_value(u.AA) * (1. + vi / c),
                #     self.flux[i, :],
                #     fill=None,
                # ) << self.flux.unit
                fac = 1.0 + vi / c
                new_axis[:] = ref_axis[:] * fac
                new_flux[i, :] = np.interp(
                    ref_axis,
                    new_axis,
                    self.flux[i, :],
                )

            out = Spectra(
                spectral_axis=self.spectral_axis,
                flux=new_flux << self.flux.unit,
                meta=copy(self.meta),
            )
            return out

    def resample(self, new_wave: u.Quantity):
        """
        Resample the spectra


        Parameters
        ----------
        new_wave
            Spectral axis with the desired sampling for the spectra. These values
            are the positions of the bin centers.

        Notes
        -----
        This functions makes use of :func:`spectres.spectres` [1]_.
        Care must be taken when resampling near the boundaries of the spectral
        axis, which can cause `nan` values.

        .. [1] A. C. Carnall, "SpectRes: A Fast Spectral Resampling Tool in Python",
           arXiv:1705.05165

        """
        new_flux = spectres(
            new_wave.to_value(u.AA),
            self.spectral_axis.to_value(u.AA),
            self.flux,
            fill=None,
        )
        out = Spectra(
            spectral_axis=new_wave,
            flux=new_flux << self.flux.unit,
            meta=copy(self.meta),
        )
        return out

    def convolve(self, lsf: u.Quantity = u.Quantity(1, unit=u.AA), lsf_wave=None):
        """
        Returns a convolved version of the spectra. It does the convolution
        using the FWHM given in the input line spread function (LSF).

        Notes
        -----
        If output LSF < input LSF the sigma used for the convolution is set to
        very small values.

        Parameters
        ----------
        lsf: `~astropy.units.Quantity`
            Line spread function as a function of `lsf_wave`. This is the
            FWMH to be used in the convolution, thus, should be provide in units
            of wavelength. It accepts a scalar value, that is assumend constant
            for all wavelenghts.
        lsf_wave: `~astropy.units.Quantity`
            Associated wavelenghts to the values of `lsf`.

        Returns
        -------
        Spectra
            Object instance with convolved spectra

        """

        logger.info("Convolving spectra")

        if np.isscalar(lsf.to_value(u.AA)):
            out_lsf = np.full(self.npix, lsf) * lsf.unit
        else:
            out_lsf = np.interp(self.spectral_axis, lsf_wave, lsf)

        # In most cases this interpolation is trivial, but allows the
        # flexibility to do the convolution after trimming the spectra
        in_lsf = np.interp(
            self.spectral_axis, self.meta["lsf_wave"], self.meta["lsf_fwhm"]
        )

        # NB: there is a factor 2.355 between the FWHM and the \sigma of a gaussian
        sigma = np.sqrt(out_lsf**2 - in_lsf**2) / 2.355
        bad = np.isnan(sigma)
        sigma[bad] = 1e-10 * u.AA
        out_lsf[bad] = in_lsf[bad]

        if self.nspec == 1:
            outflux = Spectra._gaussian_filter1d(
                self.flux.value, sigma.to_value(self.spectral_axis.unit)
            )
        else:
            outshape = self.flux.shape
            outflux = np.empty(outshape)
            for index in np.ndindex(outshape[:-1]):
                flux = Spectra._gaussian_filter1d(self.flux[index].value, sigma)
                s = slice(None)
                outflux[index + (s,)] = flux

        out = Spectra(
            spectral_axis=self.spectral_axis,
            flux=outflux * self.flux.unit,
            meta=copy(self.meta),
        )
        out.meta["lsf_fwhm"] = out_lsf
        out.meta["lsf_wave"] = self.spectral_axis
        return out

    # -----------------------------------------------------------------------------
    def tune_spectra(
        self,
        wave_lims=None,
        dwave=None,
        sampling=None,
        redshift=None,
        lsf_flag=False,
        lsf_mode="FWHM",
        lsf_wave=None,
        lsf=None,
    ):
        # This wrapper should be defined at the flask application layer
        """
        Returns the a tuned to desired input parameters

        Parameters
        ----------
        wave_lims:
            Wavelength limits in Angstroms
        dwave:
            Step in wavelength (in Angstroms)
        sampling:
            Type of sampling of the spectra. Valid inputs are lin/ln.
            Default: lin
        redshift:
            Desired redshift
        lsf_flag:
            Boolean flag to do LSF correction
        lsf_wave:
            Wavelength vector of output LSF
        lsf:
            LSF vector
        lsf_mode:
            FWHM/VDISP. First one in Angstroms. Second one in km/s

        Returns
        -------
        Spectra
            Object instance with tuned spectra and updated info

        """

        logger.debug("Tuning spectra ----------------------")

        out = copy(self)
        if lsf_wave is None:
            lsf_wave = out.wave

        # Resampling the spectra if necessary
        # to be done with specutils
        if (
            (wave_lims[0] != out.wave_init)
            or (wave_lims[1] != out.wave_last)
            or (dwave != self.dwave)
        ):
            pass
        #    out = self.resample_spectra(wave_lims=wave_lims, dwave=dwave)

        # Redshift spectra if necessary
        # to be done with specutils
        if redshift != out.redshift:
            pass
        #    out = out.redshift_spectra(redshift=redshift)

        # Convolving spectra is necessary
        if lsf is not None:
            out = out.convolve(lsf_wave=lsf_wave, lsf=lsf, mode="FWHM")

        # Log-rebinning spectra if necessary
        # is this really needed?
        if sampling == "ln":
            out = out.logrebin_spectra()

        return out

    def magnitudes(
        self,
        filters: list[Filter] = [],
        zeropoint="AB",
    ) -> Magnitude:
        """
        Returns the magnitudes of the input spectra given a list of filters in a file

        Parameters
        ----------
        filters: list[Filter]
            Filters as provided by :meth:`milespy.filter.get`
        zeropoint:
            Type of zero point. Valid inputs are AB/VEGA

        Returns
        -------
        Magnitude
            Dictionary with output magnitudes for each spectra for each filter

        """
        logger.info("Computing absolute magnitudes")

        outmags = compute_mags(self.spectral_axis, self.flux, filters, zeropoint)

        return outmags

    # -----------------------------------------------------------------------------
    def line_strength(self, indeces: list[LineStrengthIndex]) -> LineStrengthDict:
        """
        Returns the LS indices of the input spectra given a list of index definitions

        Parameters
        ----------
        indeces: list[LineStrenghtIndex]
            Indeces as provided by :meth:`milespy.ls_indices.get`

        Returns
        -------
        LineStrengthDict
            Dictionary with output LS indices for each spectra for each index

        """
        logger.info("Computing Line-Strength indices")
        outls = lsindex(
            indeces,
            self.spectral_axis,
            self.flux,
            self.redshift,
        )

        """
        nls = len(indices)
        indices = np.zeros((nls, self.nspec))

        for i in tqdm(range(self.nspec), delay=3.):
            names, indices[:, i], dummy = lsindex(
                self.spectral_axis,
                self.flux,
                self.flux * 0.1,
                self.redshift,
                0.0,
                self.lsfile,
            )

        outls = LineStrengthDict((names[i], indices[i, :]) for i in range(nls))
        """

        return outls

    @staticmethod
    def _gaussian_filter1d(spec, sig):
        """
        Convolve a spectrum by a Gaussian with different sigma for every pixel.

        If all sigma are the same this routine produces the same output as
        scipy.ndimage.gaussian_filter1d, except for the border treatment.
        Here the first/last p pixels are filled with zeros.
        When creating a template library for SDSS data, this implementation
        is 60x faster than a naive for loop over pixels.

        Parameters
        ----------
        spec:
            vector with the spectrum to convolve
        sig:
            vector of sigma values (in pixels) for every pixel

        Returns
        -------
        spec: ndarray
            Spectrum convolved with a Gaussian with dispersion sig

        """
        sig = sig.clip(0.01)  # forces zero sigmas to have 0.01 pixels
        p = int(np.ceil(np.max(3 * sig)))
        m = 2 * p + 1  # kernel size
        x2 = np.linspace(-p, p, m) ** 2

        n = spec.size
        a = np.zeros((m, n))
        for j in range(m):  # Loop over the small size of the kernel
            a[j, p:-p] = spec[j : n - m + j + 1]

        gau = np.exp(-x2[:, None] / (2 * sig**2))
        gau /= np.sum(gau, 0)[None, :]  # Normalize kernel

        conv_spectrum = np.sum(a * gau, 0)

        return conv_spectrum

    def mass_to_light(
        self, filters: list[Filter], mass_in: typing.Union[str, list[str]] = "star+remn"
    ) -> dict:
        """
        Computes the mass-to-light ratios of models in the desired filters

        Parameters
        ----------
        filters: list[Filter]
            Filters as provided by the method 'get_filters"
        mass_in: str | list[str]
            What mass to take into account for the ML. It can be given as a list,
            so that it returns a dictionary for each type.
            Valid values are: total, star, remn, star+remn, gas

        Returns
        -------
        dict
            Dictionary with mass-to-light ratios for each SSP model and filter.
            If mass_in is a list, the first key is the type of ML.

        """
        logger.info("Computing mass-to-light ratios")

        if type(mass_in) is str:
            mass_in = [mass_in]

        #  We need to choose a system. For M/Ls this is irrelevant
        zeropoint = "AB"
        mags = self.magnitudes(filters=filters, zeropoint=zeropoint)
        msun = sun_magnitude(filters=filters, zeropoint=zeropoint)

        outmls = {}
        logger.debug(f"{self.meta.keys()}")
        for m in mass_in:
            if m == "total":
                mass = self.meta["Mass_total"]
            elif m == "remn":
                mass = self.meta["Mass_remn"]
            elif m == "star":
                mass = self.meta["Mass_star"]
            elif m == "star+remn":
                mass = self.meta["Mass_star_remn"]
            elif m == "gas":
                mass = self.meta["Mass_gas"]
            else:
                raise ValueError(
                    "Mass type not allowed. "
                    "Valid options are total, star, remn, star+remn, gas"
                )

            outmls[m] = self._single_type_mass_to_light(filters, mass, mags, msun)

        # If only a single mass is requested we omit the information in the
        # returned dictionary
        if len(mass_in) == 1:
            return outmls[mass_in[0]]
        else:
            return outmls

    def _single_type_mass_to_light(self, filters: list[Filter], mass, mags, msun):
        outmls = {}
        for f in filters:
            outmls[f.name] = (mass / 1.0) * 10 ** (
                -0.40 * (msun[f.name] - mags[f.name])
            )
        return outmls

    def _update_mass(self, mass):
        if "mass" in self.meta.keys():
            previous_mass = self.meta["mass"]
        else:
            previous_mass = np.ones(mass.shape) << u.Msun
            self.meta["mass"] = previous_mass

        ratio = mass.to_value(u.Msun) / previous_mass.to_value(u.Msun)
        if np.ndim(ratio) == 0:
            if ratio <= 0.0:
                raise ValueError("Trying to set zero or negative mass to an spectra")
        else:
            # This can happens for spectra in a SFH that have a total contribution
            # of zero
            ratio[np.isnan(ratio)] = 0.0

        self.meta["mass"] *= ratio
        self.meta["Mass_total"] *= ratio
        self.meta["Mass_remn"] *= ratio
        self.meta["Mass_star"] *= ratio
        self.meta["Mass_star_remn"] *= ratio
        self.meta["Mass_gas"] *= ratio

    def _apply_mass(self, mass):
        if np.ndim(mass) == 0:
            m = mass
        else:
            m = mass[:, np.newaxis]

        out = self.multiply(m, handle_meta="first_found")
        out._update_mass(mass)
        return out



# -*- coding: utf-8 -*-
from __future__ import annotations

import logging
import re
import sys
from typing import Literal

import matplotlib.pyplot as plt
import numpy as np
from astropy import units as u
from astropy.io import ascii

from .configuration import get_config_file

logger = logging.getLogger("milespy.ls_indices")

lsfile = get_config_file("ls_indices_full.def")
lsindex_names = ascii.read(lsfile, comment=r"\s*#")["names"]
logging.debug(f"Initialized line strength with {len(lsindex_names)} indeces.")


class LineStrengthIndex:
    """
    Line strength index information

    Attributes
    ----------
    name: str
        Name of the index.
    index_type: str
        Whether the index is for atomic ("A") or molecular ("M") lines.
    band_blue1:
        Left edge of the blue band
    band_blue2:
        Right edge of the blue band
    band_centre1:
        Left edge of the central bandpass
    band_centre2:
        Right edge of the central bandpass
    band_red1:
        Left edge of the red band
    band_red2:
        Right edge of the red band
    """

    def __init__(
        self,
        name: str,
        index_type: Literal["A", "M"],
        band_blue1: u.Quantity[u.AA],
        band_blue2: u.Quantity[u.AA],
        band_centre1: u.Quantity[u.AA],
        band_centre2: u.Quantity[u.AA],
        band_red1: u.Quantity[u.AA],
        band_red2: u.Quantity[u.AA],
    ):
        self.name = name
        self.type = index_type

        self.bands = np.empty(6)
        self.bands[0] = band_blue1.to(u.AA).value
        self.bands[1] = band_blue2.to(u.AA).value
        self.bands[2] = band_centre1.to(u.AA).value
        self.bands[3] = band_centre2.to(u.AA).value
        self.bands[4] = band_red1.to(u.AA).value
        self.bands[5] = band_red2.to(u.AA).value

    @staticmethod
    def from_database(name) -> LineStrengthIndex:
        """
        Create an index from the name in the database.

        This reads the information from the configuration files, so the
        name should match with a given existing index. This can be easily
        accomplished with :meth:`milespy.index.search`.

        Parameters
        ----------
        name : str
            Name of the index to be loaded

        Raises
        ------
        ValueError
            If there is no matching index in the database
        RuntimeError
            If there are multiple indeces that match the input name
        """
        tab = ascii.read(lsfile, comment=r"\s*#")
        names = tab["names"]
        if name in names:
            idx = np.argwhere(names == name)
            if len(idx) > 1:
                raise RuntimeError("Multiple matching filters")

            return LineStrengthIndex(
                name,
                tab["b7"][idx][0][0],
                tab["b1"][idx][0][0] << u.AA,
                tab["b2"][idx][0][0] << u.AA,
                tab["b3"][idx][0][0] << u.AA,
                tab["b4"][idx][0][0] << u.AA,
                tab["b5"][idx][0][0] << u.AA,
                tab["b6"][idx][0][0] << u.AA,
            )

        else:
            raise ValueError(f"The index {name} is not on the database")


def search(name) -> list[str]:
    """
    Searches for an index in database.

    Notes
    -----
    Search is case insensitive
    The filter seach does not have to be precise. Substrings within filter
    names are ok.  It uses the python package 're' for regular expressions
    matching

    Parameters
    ----------
    name:
        The search string to match index names

    Returns
    -------
    list[str]
        List of index names available matching the search string

    """

    reg = re.compile(name, re.IGNORECASE)
    filtered_filters = list(filter(reg.search, lsindex_names))

    if len(filtered_filters) == 0:
        logger.warning(
            "Cannot find filter in our database\n Available filters are:\n\n"
            + str(lsindex_names)
        )

    return filtered_filters


def get(lsindex_names: list[str]) -> list[LineStrengthIndex]:
    """
    Retrieves index from database

    Parameters
    ----------
    lsindex_names: list[str]
        The index names as given by :meth:`milespy.ls_indeces.search`

    Returns
    -------
    list[Filter]
    """
    indeces = [LineStrengthIndex.from_database(name) for name in lsindex_names]

    return indeces


class LineStrengthDict(dict):
    def write(self, output=sys.stdout, format="basic", **kwargs):
        """
        Save the line strength indices in the requested format

        Any extra keyword parameters are passed to astropy.io.ascrii.write

        Parameters
        ----------
        output: str
            Output filename. Defaults to sys.stdout
        format : str
            Any of the available format specifier for astropy.io.ascii.write:
            https://docs.astropy.org/en/stable/api/astropy.io.ascii.write.html#astropy.io.ascii.write
        """
        from astropy.table import Table

        tab = Table(data=dict(self))
        ascii.write(tab, output, format=format, **kwargs)


def _sum_counts(ll, c, b1, b2):
    # Central full pixel range
    dw = ll[1] - ll[0]  # linear step size
    w = (ll >= b1 + dw / 2.0) & (ll <= b2 - dw / 2.0)
    s = np.sum(c[..., w], axis=-1)

    # First fractional pixel
    pixb = (ll < b1 + dw / 2.0) & (ll > b1 - dw / 2.0)
    if np.any(pixb):
        fracb = ((ll[pixb] + dw / 2.0) - b1) / dw
        s = s + c[..., pixb][..., 0] * fracb

    # Last fractional pixel
    pixr = (ll < b2 + dw / 2.0) & (ll > b2 - dw / 2.0)
    if np.any(pixr):
        fracr = (b2 - (ll[pixr] - dw / 2.0)) / dw
        s = s + c[..., pixr][..., 0] * fracr

    return s


def _calc_index(bands, index_type, name, ll, counts, plot=False):
    cb = _sum_counts(ll, counts, bands[0], bands[1])
    cr = _sum_counts(ll, counts, bands[4], bands[5])
    s = _sum_counts(ll, counts, bands[2], bands[3])

    lb = (bands[0] + bands[1]) / 2.0
    lr = (bands[4] + bands[5]) / 2.0
    cb = cb / (bands[1] - bands[0])
    cr = cr / (bands[5] - bands[4])
    m = (cr - cb) / (lr - lb)
    c1 = (m * (bands[2] - lb)) + cb
    c2 = (m * (bands[3] - lb)) + cb
    cont = 0.5 * (c1 + c2) * (bands[3] - bands[2])

    if index_type == "A":
        # atomic index
        ind = (1.0 - (s / cont)) * (bands[3] - bands[2])
    elif index_type == "M":
        # molecular index
        ind = -2.5 * np.log10(s / cont)

    if plot:
        minx = bands[0] - 0.05 * (bands[5] - bands[0])
        maxx = bands[5] + 0.05 * (bands[5] - bands[0])
        miny = np.amin(counts) - 0.05 * (np.amax(counts) - np.amin(counts))
        maxy = np.amax(counts) + 0.05 * (np.amax(counts) - np.amin(counts))
        plt.figure()
        plt.plot(ll, counts, "k")
        plt.xlabel(r"Wavelength ($\AA$)")
        plt.ylabel("Counts")
        plt.title(name)
        plt.xlim([minx, maxx])
        plt.ylim([miny, maxy])
        dw = ll[1] - ll[0]
        plt.plot([lb, lr], [c1 * dw, c2 * dw], "r")
        good = (ll >= bands[2]) & (ll <= bands[3])
        ynew = np.interp(ll, [lb, lr], [c1[0] * dw, c2[0] * dw])
        plt.fill_between(ll[good], counts[good], ynew[good], facecolor="green")
        for i in range(len(bands)):
            plt.plot([bands[i], bands[i]], [miny, maxy], "k--")
        plt.show()

    return ind


def lsindex(
    indeces: [LineStrengthIndex],
    ll,
    flux,
    z,
    plot=False,
    noise=None,
    z_err=None,
    sims=0,
):
    """
    Measure line-strength indices

    Author: J. Falcon-Barroso

    Parameters
    ----------
    indeces : [LineStrengthIndex]
        indeces to be computed
    ll : np.ndarray
        wavelength vector; assumed to be in *linear steps*
    flux : np.ndarray
        counts as a function of wavelength
    z : float
        redshift (in km/s)
    plot : bool
        plot spectra
    noise :
        noise spectrum
    z_err : float
        redshift error (in km/s)
    sims : int
        number of simulations for the errors (default: 100)

    Returns
    -------
    LineStrengthDict
    """
    # TODO: take into account the units
    ll = ll.to_value()
    z = z.to_value()
    flux = flux.to_value()

    # Deredshift spectrum to rest wavelength
    dll = ll / (z + 1.0)

    outindex = LineStrengthDict(
        (ind.name, np.full(flux.shape[:-1], np.nan)) for ind in indeces
    )

    for ind in indeces:
        good = (ind.bands[0] >= dll[0]) & (ind.bands[5] <= dll[-1])
        if not good:
            logger.warning(
                f"Index {ind.name} [{ind.bands[0]}, {ind.bands[5]}] "
                f"is outside of the spectral range [{dll[0]}, {dll[-1]}]"
            )
            continue
        # calculate index value
        outindex[ind.name] = _calc_index(ind.bands, ind.type, ind.name, dll, flux, plot)

    if sims > 0:
        raise NotImplementedError
        """
        # Calculate errors
        index_error = np.zeros(num_ind, dtype="D") * np.nan
        index_noise = np.zeros([num_ind, sims], dtype="D")

        # Create redshift and sigma errors
        dz = np.random.randn(sims) * z_err

        # Loop through the simulations
        for i in range(sims):
            # resample spectrum according to noise
            ran = np.random.normal(0.0, 1.0, len(dll))
            flux_n = flux + ran * noise

            # loop through all indices
            for k, ind in enumerate(indeces):
                # shift bands according to redshift error
                sz = z + dz[i]
                dll = ll / (sz + 1.0)
                bands2 = ind.bands
                if (dll[0] <= bands2[0]) and (dll[len(dll) - 1] >= bands2[5]):
                    tmp = _calc_index(bands2, ind.name, dll, flux_n, 0)
                    index_noise[k, i] = tmp
                else:
                    # index outside wavelength range
                    index_noise[k, i] = -999

        # Get STD of distribution (index error)
        index_error = np.std(index_noise, axis=1)
        """

    return outindex



# -*- coding: utf-8 -*-
from __future__ import annotations

import glob
import logging
import os
import re

import matplotlib.pyplot as plt
import numpy as np
from astropy.io import ascii

from .configuration import config_folder
from .configuration import get_config_file

logger = logging.getLogger("milespy.filter")


class Filter:
    """
    Filter information

    Attributes
    ----------
    wave: array
        Wavelength range of the filter
    trans: array
        Transmissivity for each wavelength
    name: str
        Name of the filter
    """

    def __init__(self, fname):
        """
        Create a filter from the name in the database.

        This reads the information from the configuration files, so the
        name should match with a given existing file. This can be easily
        accomplished with :meth:`milespy.filter.search`.

        Parameters
        ----------
        fname : str
            Name of the filter to be loaded
        """
        filename = get_config_file("filters/" + fname + ".dat")
        if not os.path.exists(filename):
            assert ValueError("Filter " + fname + " does not exist in database")
        else:
            tab = ascii.read(filename, names=["wave", "trans"])
            tab["trans"] /= np.amax(tab["trans"])
            self.wave = tab["wave"]
            self.name = fname
            self.trans = tab["trans"]

    def plot(self, ax) -> None:
        """
        Plot the filter transmissivity

        Parameters
        ----------
        ax : matplotlib.Axes
            Axes where the plot is drawn
        """
        ax.fill_between(
            self.wave,
            self.trans,
            alpha=0.5,
            label=self.name,
            edgecolor="k",
        )


fnames = glob.glob(f"{config_folder.as_posix()}/filters/*.dat")
filter_names = np.sort([os.path.basename(x).split(".dat")[0] for x in fnames])
nfilters = len(filter_names)
logging.debug(f"Initialized library with {nfilters} filters")


def search(name) -> list[str]:
    """
    Searches for a filter in database.

    Notes
    -----
    Search is case insensitive
    The filter seach does not have to be precise. Substrings within filter
    names are ok.  It uses the python package 're' for regular expressions
    matching

    Parameters
    ----------
    name:
        The search string to match filter names

    Returns
    -------
    list[str]
        List of filter names available matching the search string

    """

    reg = re.compile(name, re.IGNORECASE)
    filtered_filters = list(filter(reg.search, filter_names))

    if len(filtered_filters) == 0:
        logger.warning(
            "Cannot find filter in our database\n Available filters are:\n\n"
            + str(filter_names)
        )

    return filtered_filters


def get(filter_names: list[str]) -> list[Filter]:
    """
    Retrieves filter from database

    Parameters
    ----------
    filter_names: list[str]
        The filter names as given by :meth:`milespy.filter.search`

    Returns
    -------
    list[Filter]
    """
    filters = [Filter(fname) for fname in filter_names]

    return filters


def plot(filter_names, legend=True):
    """
    Plot filters

    Parameters
    ----------
    filter_names: list[str]
        The filter names
    legend: bool
        Flag to turn on/off the legend

    """

    fig, ax = plt.subplots()

    _ = [Filter(fname).plot(ax) for fname in filter_names]

    if legend:
        plt.legend()

    plt.show()



# -*- coding: utf-8 -*-
import logging
from typing import Optional

import h5py
import numpy as np
import numpy.typing as npt
from astropy import units as u
from astropy.units import Quantity
from scipy.spatial import Delaunay
from tqdm import tqdm

from .misc import interp_weights
from .repository import Repository
from .spectra import Spectra


logger = logging.getLogger("milespy.stellarlib")


class StellarLibrary(Repository):
    """
    Single stars library.

    Attributes
    ----------
    models: Spectra
        Spectra of all the stars that form the loaded library
    source: str
        Name of input library being used
    version: str
        Version number of the library
    """

    def __init__(self, source="MILES_STARS", version="9.1"):
        """
        Creates an instance of the class

        Parameters
        ----------
        source:
            Name of input models to use.
            If none of the options below, it will assume that `source` is the path
            to an external model repository file.
            Valid inputs are
            MILES_STARS/CaT_STARS/EMILES_STARS
        version:
            Version number of the models

        """
        repo_filename = self._get_repository(source, version)
        self._assert_repository_file(repo_filename)

        # Opening the relevant file in the repository
        f = h5py.File(repo_filename, "r")
        # ------------------------------
        meta = {
            "index": np.array(f["index"]),
            "teff": np.array(f["teff"]) << u.K,
            "logg": np.array(f["logg"]) << u.dex,
            "FeH": np.array(f["FeH"]) << u.dex,
            "MgFe": np.array(f["MgFe"]) << u.dex,
            "starname": np.array([n.decode() for n in f["starname"]]),
            "filename": np.array([n.decode() for n in f["filename"]]),
            "id": np.array([np.int32(n.decode()) for n in f["id"]]),
        }
        wave = np.array(f["wave"])
        spec = np.array(f["spec"])
        self.source = source
        self.version = version
        # ------------------------------
        f.close()

        # Flagging if all elements of MgFe are NaNs
        self.MgFe_nan = False
        if np.nansum(meta["MgFe"]) == 0:
            self.MgFe_nan = True

        # Creating Delaunay triangulation of parameters for future searches and
        # interpolations
        if self.MgFe_nan:
            idx = (
                np.isfinite(meta["teff"])
                & np.isfinite(meta["logg"])
                & np.isfinite(meta["FeH"])
            )
            ngood = np.sum(idx)
            self.params = np.empty((ngood, 3))
            self.params[:, 0] = np.log10(meta["teff"].to_value(u.K))[idx]
            self.params[:, 1] = meta["logg"][idx]
            self.params[:, 2] = meta["FeH"][idx]
        else:
            idx = (
                np.isfinite(meta["teff"])
                & np.isfinite(meta["logg"])
                & np.isfinite(meta["FeH"])
                & np.isfinite(meta["MgFe"])
            )
            ngood = np.sum(idx)
            self.params = np.empty((ngood, 4))
            self.params[:, 0] = np.log10(meta["teff"])[idx]
            self.params[:, 1] = meta["logg"][idx]
            self.params[:, 2] = meta["FeH"][idx]
            self.params[:, 3] = meta["MgFe"][idx]

        self.tri = Delaunay(self.params)
        self.index = meta["index"][idx]
        self.main_keys = list(self.__dict__.keys())

        super().__init__(
            Spectra(
                spectral_axis=Quantity(wave, unit=u.AA),
                flux=Quantity(spec.T, unit=None),
                meta=meta,
            )
        )

    def search_by_id(self, id=None):
        """
        Searches a star in database for a given ID

        Parameters
        ----------
        id:
            integer with the star ID in database

        Returns
        -------
        Object instance for selected items

        """

        idx = self._id_to_idx(id)

        out = Spectra.__getitem__(self.models, idx)

        return out

    def _id_to_idx(self, lib_id: npt.ArrayLike) -> np.ndarray:
        id_arr = np.array(lib_id, ndmin=1)
        common, _, idx = np.intersect1d(
            id_arr, self.models.meta["id"], assume_unique=True, return_indices=True
        )
        if len(common) != len(id_arr):
            raise ValueError("No star with that ID")
        return idx

    # -----------------------------------------------------------------------------

    def get_starname(self, id=None):
        """
        Gets a starname in database for a given ID

        Parameters
        ----------
        id:
            integer with the star ID in database

        Returns
        -------
        Star name

        """

        idx = self._id_to_idx(id)
        logger.debug(f"{type(idx[0])}")

        return self.models.meta["starname"][idx]

    # -----------------------------------------------------------------------------

    @u.quantity_input
    def in_range(
        self,
        teff_lims: u.Quantity[u.K],
        logg_lims: u.Quantity[u.dex],
        FeH_lims: u.Quantity[u.dex],
        MgFe_lims: Optional[u.Quantity[u.dex]] = None,
    ):
        """
        Gets set of stars with parameters range

        Parameters
        ----------
        teff_lims:
            Limits in Teff
        logg_lims:
            Limits in Log(g)
        FeH_lims:
            Limits in [Fe/H]
        MgFe_lims:
            Limits in [Mg/Fe]

        Returns
        -------
        StellarLibrary
            Object instance for stars within parameters range

        """

        if self.MgFe_nan:
            idx = (
                (self.models.meta["teff"] >= teff_lims[0])
                & (self.models.meta["teff"] <= teff_lims[1])
                & (self.models.meta["logg"] >= logg_lims[0])
                & (self.models.meta["logg"] <= logg_lims[1])
                & (self.models.meta["FeH"] >= FeH_lims[0])
                & (self.models.meta["FeH"] <= FeH_lims[1])
            )
        else:
            idx = (
                (self.models.meta["teff"] >= teff_lims[0])
                & (self.models.meta["teff"] <= teff_lims[1])
                & (self.models.meta["logg"] >= logg_lims[0])
                & (self.models.meta["logg"] <= logg_lims[1])
                & (self.models.meta["FeH"] >= FeH_lims[0])
                & (self.models.meta["FeH"] <= FeH_lims[1])
                & (self.models.meta["MgFe"] >= MgFe_lims[0])
                & (self.models.meta["MgFe"] <= MgFe_lims[1])
            )

        out = Spectra.__getitem__(self.models, idx)

        return out

    @u.quantity_input
    def closest(
        self,
        teff: u.Quantity[u.K],
        logg: u.Quantity[u.dex],
        FeH: u.Quantity[u.dex],
        MgFe: Optional[u.Quantity[u.dex]] = None,
    ):
        """
        Gets closest star in database for given set of parameters

        Parameters
        ----------
        teff:
            Desired Teff
        logg:
            Desired Log(g)
        FeH:
            Desired [Fe/H]
        MgFe:
            Desired [Mg/Fe]

        Returns
        -------
        Spectra
            Spectrum from the closest star in the library.

        """
        return self.interpolate(teff, logg, FeH, MgFe, closest=True)

    @u.quantity_input
    def interpolate(
        self,
        teff: u.Quantity[u.K],
        logg: u.Quantity[u.dex],
        FeH: u.Quantity[u.dex],
        MgFe: Optional[u.Quantity[u.dex]] = None,
        closest: bool = False,
        simplex: bool = False,
    ):
        """
        Interpolates a star spectrum for given set of parameters using Delaunay
        triangulation

        Parameters
        ----------
        teff:
            Desired Teff
        logg:
            Desired Log(g)
        FeH:
            Desired [Fe/H]
        MgFe:
            Desired [Mg/Fe]
        closest:
            Return the closest spectra, rather than performing the interpolation.
            If only one interpolation is performed, all the spectra in the simplex
            vertices are returned.
        simplex:
            If only one set of input parameters is given, return all the spectra
            that form part of the simplex used for the interpolation. These spectra
            have the weights information in their `meta` dictionary.

        Returns
        -------
        Spectra
            Interpolated spectrum.
            If closest == True, return the closest spectra from the repository,
            rather than doing the interpolation.

        Raises
        ------
        RuntimeError
            If the values are out of the grid.
        ValueError
            If the provided parameters do not have the same shape.
        """

        teff = np.atleast_1d(teff)
        logg = np.atleast_1d(logg)
        FeH = np.atleast_1d(FeH)
        MgFe = np.atleast_1d(MgFe)

        wrong_shape = teff.shape != logg.shape
        wrong_shape |= teff.shape != FeH.shape
        if not self.MgFe_nan:
            wrong_shape |= teff.shape != MgFe.shape
        if wrong_shape:
            raise ValueError("The input parameters should all have the same shape")

        ninterp = len(teff)

        # Checking input point is within the grid
        def _compute_bounds(val, model):
            min_model = np.amin(model)
            max_model = np.amax(model)
            if np.any(val < min_model) or np.any(val > max_model):
                raise RuntimeError(
                    f"Input parameters {val} is outside of model grid: "
                    f"({min_model},{max_model})"
                )

        _compute_bounds(teff, self.models.meta["teff"])
        _compute_bounds(logg, self.models.meta["logg"])
        _compute_bounds(FeH, self.models.meta["FeH"])
        if not self.MgFe_nan:
            _compute_bounds(MgFe, self.models.meta["MgFe"])

        if closest:
            closest_idx = np.empty(ninterp, dtype=int)
        else:
            wave = self.models.spectral_axis
            spec = Quantity(
                value=np.empty((ninterp, self.models.data.shape[1])),
                unit=self.models.flux.unit,
            )
            new_meta = {
                "teff": teff,
                "logg": logg,
                "FeH": FeH,
                "MgFe": MgFe,
            }
            base_keys = list(new_meta.keys())
            for k in self.models.meta.keys():
                if k not in new_meta.keys():
                    if len(self.models.meta[k]) > 1:
                        if "U" not in self.models.meta[k].dtype.kind:
                            new_meta[k] = np.empty(
                                ninterp, dtype=self.models.meta[k].dtype
                            )

        for i in tqdm(range(ninterp), delay=3.0):
            if self.MgFe_nan:
                input_pt = np.atleast_2d(
                    [np.log10(teff[i].to_value(u.K)), logg[i].value, FeH[i].value]
                )
            else:
                input_pt = np.atleast_2d(
                    [
                        np.log10(teff[i].to_value(u.K)),
                        logg[i].value,
                        FeH[i].value,
                        MgFe[i].value,
                    ]
                )

            vtx, wts = interp_weights(self.params, input_pt, self.tri)
            vtx, wts = vtx.ravel(), wts.ravel()

            if logger.isEnabledFor(logging.DEBUG):
                logger.debug(f"Simplex ids: {self.models.meta['index'][vtx]}")
                logger.debug(f"Simple ages: {self.models.meta['teff'][vtx]}")
                logger.debug(f"Simplex met: {self.models.meta['logg'][vtx]}")
                logger.debug(f"Simplex met: {self.models.meta['FeH'][vtx]}")
                logger.debug(f"Simplex weights: {wts}, norm: {np.sum(wts)}")

            idx = self.index[vtx]

            if closest:
                if simplex and ninterp == 1:
                    out = Spectra.__getitem__(
                        self.models, self.models.meta["index"][idx]
                    )
                    return out
                else:
                    closest_idx[i] = idx[np.argmax(wts)]
            else:
                spec[i, :] = np.dot(self.models.flux[idx, :].T, wts)

                # Interpolate the rest of the meta if possible
                for k in new_meta:
                    if len(self.models.meta[k]) > 1:
                        if k not in base_keys:
                            new_meta[k] = np.dot(self.models.meta[k][idx], wts)
        if closest:
            out = Spectra.__getitem__(self.models, closest_idx)
        else:
            out = Spectra(spectral_axis=wave, flux=spec, meta=new_meta)
        # Select the first and only spectra so that the users does not need to
        # do this all the time, manually
        if ninterp == 1:
            return out[0]
        else:
            return out



# -*- coding: utf-8 -*-
import logging
import os

import astropy.units as u
import h5py
import requests
from specutils.manipulation import spectral_slab
from tqdm import tqdm

from .configuration import config
from .configuration import def_repo_folder

logger = logging.getLogger("milespy.repository")


repository_url = {
    "MILES_STARS_v9.1": "https://cloud.iac.es/index.php/s/TKEwKfSiaZePYsx/download/MILES_STARS_v9.1.hdf5",  # noqa
    "MILES_SSP_v9.1": "https://cloud.iac.es/index.php/s/wz3xS9jj7zDe7Hs/download/MILES_SSP_v9.1.hdf5",  # noqa
    "EMILES_SSP_v9.1": "https://cloud.iac.es/index.php/s/2CqEBsreXdeK2Pd/download/EMILES_SSP_v9.1.hdf5",  # noqa
    "CaT_STARS_v9.1": "https://cloud.iac.es/index.php/s/jCt2TzD8DMFXXdZ/download/CaT_STARS_v9.1.hdf5",  # noqa
    "CaT_SSP_v9.1": "https://cloud.iac.es/index.php/s/ex3Ep9jA5eG6Pwt/download/CaT_SSP_v9.1.hdf5",  # noqa
}


class Repository:
    def __init__(self, models):
        self._models = models

    def _assert_repository_file(self, file_path):
        try:
            with h5py.File(file_path) as f:
                _ = f["wave"]
        except:  # noqa
            raise AssertionError("Repository file is unreadable")

    def _download_repository(self, base_name, output_path):
        response = requests.get(repository_url[base_name], stream=True)

        total_size = int(response.headers.get("content-length", 0))
        block_size = 1024

        with tqdm(total=total_size, unit="B", unit_scale=True) as progress_bar:
            with open(output_path, "wb") as file:
                for data in response.iter_content(block_size):
                    progress_bar.update(len(data))
                    file.write(data)

        if total_size != 0 and progress_bar.n != total_size:
            raise RuntimeError("Unable to download file")

        logger.debug(f"Dowloaded {base_name} repository in {output_path}")

    def _get_repository(self, source, version) -> str:
        base_name = source + "_v" + version
        if "repository_folder" in config:
            repo_filename = config["repository_folder"] + base_name + ".hdf5"
        else:
            repo_filename = def_repo_folder.as_posix() + "/" + base_name + ".hdf5"

        logger.debug(f"Loading models in {repo_filename}")

        if base_name in repository_url.keys():
            if not os.path.exists(repo_filename):
                logger.warning("Unable to locate repository")
                if "auto_download" in config.keys() and config["auto_download"]:
                    self._download_repository(base_name, repo_filename)
                else:
                    opt = input(
                        f"Do you want to download the {base_name} repository? [y/n]: "
                    )
                    if opt == "y":
                        self._download_repository(base_name, repo_filename)
        else:
            logger.debug(f"Not known URL for {base_name}, trying to load it as a file")
            return source

        return repo_filename

    @property
    def models(self):
        return self._models

    def trim(self, lower: u.Quantity, upper: u.Quantity):
        """
        Trim all the models in the library

        Parameters
        ----------
        lower : Quantity
        upper : Quantity
        """
        trimmed = spectral_slab(self.models, lower, upper)
        trimmed.meta = self.models.meta
        self._models = trimmed

    def resample(self, new_wave: u.Quantity):
        """
        Resample all the models in the library

        Parameters
        ----------
        new_wave
            Spectral axis with the desired sampling for the spectra

        See Also
        --------
        :meth:`milespy.spectra.resample`
        """
        resample = self.models.resample(new_wave)
        resample.meta = self.models.meta
        self._models = resample
